# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing

from igraph import Graph as _Graph, Vertex, VertexSeq, Edge, EdgeSeq


VERTEX_PARAMETER_TYPE = typing.Union[Vertex, str, int]


class Graph:

    def __init__(self, directed: bool, name: str):

        self._graph: _Graph = _Graph(directed=directed, graph_attrs={r'name': name})
        self._vertices: typing.Dict[str, Vertex] = {}

    def __len__(self):

        return len(self._vertices)

    def __repr__(self):

        return f'<DiGraph: {self.name}, vertices: {self._graph.vcount()}, edges: {self._graph.ecount()}>'

    @property
    def vertices(self) -> typing.Dict[str, Vertex]:

        return self._vertices

    @property
    def name(self) -> str:

        return self._graph[r'name']

    @property
    def data(self) -> _Graph:

        return self._graph

    @property
    def vertex_seq(self) -> VertexSeq:

        return self._graph.vs

    @property
    def edge_seq(self) -> EdgeSeq:

        return self._graph.es

    def clear(self):

        self._graph.clear()
        self._vertices.clear()

    def find_vertex(self, name: str) -> Vertex:

        return self._vertices.get(name)

    def find_vertices(self, names: typing.List[str]) -> typing.List[Vertex]:

        return [self._vertices.get(name) for name in names if name in self._vertices]

    def add_vertex(self, name: str = None, **kwargs) -> Vertex:

        vertex = self._vertices.get(name)

        if vertex is None:
            vertex = self._vertices[name] = self._graph.add_vertex(name, **kwargs)

        return vertex

    def init_edges(self, edges: typing.List[typing.Tuple[VERTEX_PARAMETER_TYPE, VERTEX_PARAMETER_TYPE]]):

        self._graph.add_edges(edges)

    def get_edge(self, source: VERTEX_PARAMETER_TYPE, target: VERTEX_PARAMETER_TYPE) -> Edge:

        eid = self._graph.get_eid(source, target, error=False)

        if eid >= 0:
            return self._graph.es[eid]

    def add_edge(self, source: VERTEX_PARAMETER_TYPE, target: VERTEX_PARAMETER_TYPE):

        eid = self._graph.get_eid(source, target, error=False)

        if eid < 0:
            self._graph.add_edges([(source, target)])

    def del_edge(self, source: VERTEX_PARAMETER_TYPE, target: VERTEX_PARAMETER_TYPE):

        eid = self._graph.get_eid(source, target, error=False)

        if eid >= 0:
            self._graph.delete_edges(eid)

    def del_vertex_edge(self, name: str, mode=r'all'):

        vertex = self.find_vertex(name)

        if vertex is None:
            return

        if mode == r'in':
            edges = vertex.in_edges()
        elif mode == r'out':
            edges = vertex.out_edges()
        else:
            edges = vertex.all_edges()

        if edges:
            self._graph.delete_edges(edges)

    def deep_search(self, name: str, mode=r'all') -> typing.List[str]:

        vertex = self.find_vertex(name)

        if vertex is None:
            return []

        vertices = self._graph.subcomponent(vertex.index, mode=mode)

        return [self._graph.vs[_v][r'name'] for _v in vertices if _v != vertex.index]


class DiGraph(Graph):

    def __init__(self, name: str):

        super().__init__(True, name)

    def find_in(self, name: str) -> typing.List[Vertex]:

        vertex = self.find_vertex(name)

        if vertex is not None:
            return vertex.predecessors()

    def find_out(self, name: str) -> typing.List[Vertex]:

        vertex = self.find_vertex(name)

        if vertex is not None:
            return vertex.successors()

    def find_tree(self, name: str, depth=0xff, mode=r'out') -> typing.Dict[str, typing.List[str]]:

        vertex = self.find_vertex(name)

        if vertex is None:
            return {}

        nodes = {}
        vertices = [vertex]

        for _ in range(depth):

            vertices, _vertices = [], vertices

            for _vertex in _vertices:

                _sub_vertices = _vertex.successors() if mode == r'out' else _vertex.predecessors()

                if _sub_vertices:

                    _nodes = nodes[_vertex[r'name']] = []

                    for _sub_vertex in _sub_vertices:

                        if _sub_vertex[r'name'] not in nodes:
                            vertices.append(_sub_vertex)

                        _nodes.append(_sub_vertex[r'name'])

        return nodes



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing
import signal
import socket

import grpc
import msgpack

from abc import abstractmethod
from pydantic import BaseModel
from starlette.datastructures import Address

from ...extend.asyncio.base import Utils


UNARY_UNARY_MODE = grpc.unary_unary_rpc_method_handler
UNARY_STREAM_MODE = grpc.unary_stream_rpc_method_handler
STREAM_UNARY_MODE = grpc.stream_unary_rpc_method_handler
STREAM_STREAM_MODE = grpc.stream_stream_rpc_method_handler

GRPC_MODE_TYPE = typing.Union[
    typing.Type[UNARY_UNARY_MODE],
    typing.Type[UNARY_STREAM_MODE],
    typing.Type[STREAM_UNARY_MODE],
    typing.Type[STREAM_STREAM_MODE]
]


def get_remote_peer(metadata: grpc.aio.Metadata, context: grpc.aio.ServicerContext) -> str:
    return dict(metadata).get(r'remote-peer', context.peer())


def get_remote_addr(peer: str) -> Address:
    peer_split = peer.split(':')
    return Address(host=peer_split[1], port=int(peer_split[2]))


class StreamHandler:

    def __init__(self, request: typing.AsyncIterator, context: grpc.aio.ServicerContext):

        self._request: typing.AsyncIterator = request
        self._context: grpc.aio.ServicerContext = context

        self._context.add_done_callback(self.on_close)

        self._remote_peer: str = get_remote_peer(self._context.invocation_metadata(), context)
        self._remote_addr: Address = get_remote_addr(self._remote_peer)

    async def join(self) -> typing.Any:

        await self.on_connect()

        async for _message in self._request:
            await self.on_message(_message)

    @property
    def peer(self) -> str:
        return self._remote_peer

    @property
    def client(self) -> Address:
        return self._remote_addr

    def metadata(self) -> typing.Dict[str, str]:
        return dict(self._context.invocation_metadata())

    async def close(self):

        if not self._context.done():
            await self._context.abort(grpc.StatusCode.CANCELLED)

    async def read(self) -> typing.Any:

        data = await self._context.read()

        if data != grpc.aio.EOF:
            return data
        else:
            raise ConnectionAbortedError()

    async def send(self, message: typing.Any):

        if not self._context.done():
            await self._context.write(message)
        else:
            Utils.log.warning(f'grpc stream send canceled: {message}')

    @abstractmethod
    async def on_message(self, message: typing.Any):
        """
        接收消息回调
        """

    @abstractmethod
    def on_connect(self):
        """
        与客户端通信连接成功后回调
        """

    @abstractmethod
    def on_close(self, context: grpc.aio.ServicerContext):
        """
        会话关闭回调
        """


class Router:

    def __init__(self, service: str):

        self._service: str = service
        self._handlers: typing.Dict[str, typing.Tuple[typing.Callable, GRPC_MODE_TYPE]] = {}

    @property
    def service(self) -> str:
        return self._service

    @property
    def handlers(self) -> typing.List[typing.Tuple[typing.Callable, GRPC_MODE_TYPE]]:
        return list(self._handlers.values())

    @staticmethod
    def _get_base_model(func: typing.Callable) -> typing.Optional[typing.Type[BaseModel]]:

        annotations = typing.get_type_hints(func)

        for annotation in annotations.values():
            if issubclass(annotation, BaseModel):
                return annotation
        else:
            return None

    def register(self, func: typing.Callable, mode: GRPC_MODE_TYPE) -> typing.Callable:

        if func.__name__ in self._handlers:
            raise ValueError(f'{func.__name__} has exists')

        self._handlers[func.__name__] = (func, mode,)

        return func

    def unary_unary(self, func: typing.Callable) -> typing.Callable:

        model = self._get_base_model(func)

        @Utils.func_wraps(func)
        async def _func(request, context):
            if model is None:
                return await func(request, context)
            else:
                return await func(model(**request), context)

        return self.register(_func, UNARY_UNARY_MODE)

    def unary_stream(self, func: typing.Callable) -> typing.Callable:

        model = self._get_base_model(func)

        @Utils.func_wraps(func)
        async def _func(request, context):
            if model is None:
                return await func(request, context)
            else:
                return await func(model(**request), context)

        return self.register(_func, UNARY_STREAM_MODE)

    def stream_unary(self, func: typing.Callable) -> typing.Callable:
        return self.register(func, STREAM_UNARY_MODE)

    def stream_stream(self, func: typing.Callable) -> typing.Callable:
        return self.register(func, STREAM_STREAM_MODE)


class GRPCServer:

    def __init__(
            self,
            interceptors: typing.Optional[typing.Sequence[typing.Any]] = None,
            options: typing.Optional[grpc.aio.ChannelArgumentType] = None,
            maximum_concurrent_rpcs: typing.Optional[int] = None,
            compression: typing.Optional[grpc.Compression] = None,
            request_deserializer: typing.Optional[typing.Callable] = msgpack.loads,
            response_serializer: typing.Optional[typing.Callable] = msgpack.dumps
    ):

        self._server: grpc.aio.Server = grpc.aio.server(
            interceptors=interceptors,
            options=options,
            maximum_concurrent_rpcs=maximum_concurrent_rpcs,
            compression=compression
        )

        self._request_deserializer: typing.Optional[typing.Callable] = request_deserializer
        self._response_serializer: typing.Optional[typing.Callable] = response_serializer

        signal.signal(signal.SIGINT, self._exit)
        signal.signal(signal.SIGTERM, self._exit)

    def _exit(self, *_):
        Utils.call_soon(self.stop)

    @property
    def server(self) -> grpc.aio.Server:

        return self._server

    def register(self, service: str, handlers: typing.List[typing.Callable], *, mode: GRPC_MODE_TYPE = UNARY_UNARY_MODE):

        generic_handlers = []

        for handler in handlers:
            generic_handlers.append(
                grpc.method_handlers_generic_handler(
                    service,
                    {
                        handler.__name__: mode(
                            handler,
                            request_deserializer=self._request_deserializer,
                            response_serializer=self._response_serializer,
                        )
                    }
                )
            )

        self._server.add_generic_rpc_handlers(generic_handlers)

    def bind_router(self, router: Router):

        generic_handlers = []

        for _func, _mode in router.handlers:
            generic_handlers.append(
                grpc.method_handlers_generic_handler(
                    router.service,
                    {
                        _func.__name__: _mode(
                            _func,
                            request_deserializer=self._request_deserializer,
                            response_serializer=self._response_serializer,
                        )
                    }
                )
            )

        self._server.add_generic_rpc_handlers(generic_handlers)

    async def start(
            self,
            address: typing.Union[str, typing.Tuple[str, int]],
            family: int = socket.AF_INET,
            server_credentials: typing.Optional[grpc.ServerCredentials] = None
    ):

        if family == socket.AF_INET:
            _socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            _address = f'{address[0]}:{address[1]}'
        elif family == socket.AF_INET6:
            _socket = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
            _address = f'[{address[0]}]:{address[1]}'
        elif family == socket.AF_UNIX:
            _socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            _address = f'unix:{address}'
        else:
            raise ValueError(r'family invalid')

        _socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, True)
        _socket.bind(address)

        if server_credentials is None:
            self._server.add_insecure_port(_address)
        else:
            self._server.add_secure_port(_address, server_credentials)

        await self._server.start()

        Utils.log.success(f'grpc server [pid:{Utils.getpid()}] startup complete: {address}')

    async def stop(self, grace: typing.Optional[float] = None):
        await self._server.stop(grace)

    async def wait(self, timeout: typing.Optional[float] = None):
        return await self._server.wait_for_termination(timeout)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing
import asyncio
import threading

from tempfile import TemporaryFile

from .base import Utils
from .task import RateLimiter

from ..interface import ContextManager


class QueueBuffer:

    def __init__(
        self, handler: typing.Callable, slice_size: int, slice_time: float,
        running_limit: int, waiting_limit: int = 0
    ):

        self._lock: threading.Lock = threading.Lock()

        self._handler: typing.Callable = handler
        self._rate_limiter: RateLimiter = RateLimiter(running_limit)

        self._slice_data: asyncio.Queue[typing.Any] = asyncio.Queue(waiting_limit)
        self._slice_size: int = slice_size
        self._slice_time: float = slice_time

        self._consume_task: asyncio.Task = asyncio.create_task(self._do_consume_task())
        self._condition: asyncio.Condition = asyncio.Condition()

    def size(self) -> int:
        return self._slice_data.qsize()

    async def close(self):

        await self._slice_data.join()
        self._consume_task.cancel()

        await self._rate_limiter.close()

    async def _do_consume_task(self):

        while True:

            try:

                async with self._condition:

                    with Utils.suppress(asyncio.TimeoutError):
                        await asyncio.wait_for(
                            self._condition.wait(),
                            self._slice_time
                        )

                    while True:

                        slice_data = []

                        while not self._slice_data.empty() and len(slice_data) < self._slice_size:
                            slice_data.append(
                                self._slice_data.get_nowait()
                            )

                        if slice_data:
                            await self._rate_limiter.append(self._handler, slice_data)

                        if self._slice_data.qsize() < self._slice_size:
                            break

            except Exception as err:
                Utils.log.error(str(err))

    def safe_append(self, data: typing.Any):

        with self._lock:
            self._slice_data.put_nowait(data)

    async def append(self, data: typing.Any, timeout: typing.Union[int, float] = None):

        await asyncio.wait_for(
            self._slice_data.put(data),
            timeout
        )

        if self._slice_data.qsize() >= self._slice_size:
            if self._condition.locked():
                self._condition.notify_all()


class FileBuffer(ContextManager):
    """文件缓存类
    """

    def __init__(self, slice_size: int = 0x1000000):

        self._buffers: typing.List[TemporaryFile] = []

        self._slice_size: int = slice_size

        self._read_offset: int = 0

        self._append_buffer()

    def _context_release(self):
        self.close()

    def _append_buffer(self):
        self._buffers.append(TemporaryFile())

    def close(self):

        while len(self._buffers) > 0:
            self._buffers.pop(0).close()

        self._read_offset = 0

    def write(self, data: typing.AnyStr):

        buffer = self._buffers[-1]

        buffer.seek(0, 2)
        buffer.write(data)

        if buffer.tell() >= self._slice_size:
            buffer.flush()
            self._append_buffer()

    def read(self, size: typing.Optional[int] = None) -> typing.AnyStr:

        buffer = self._buffers[0]

        buffer.seek(self._read_offset, 0)

        result = buffer.read(size)

        if len(result) == 0 and len(self._buffers) > 1:
            self._buffers.pop(0).close()
            self._read_offset = 0
        else:
            self._read_offset = buffer.tell()

        return result



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import sys
import uvicorn

os.chdir(os.path.dirname(__file__))
sys.path.insert(0, os.path.abspath(r'../../'))

from fastapi.middleware.cors import CORSMiddleware

from hagworm.frame.fastapi.base import DEFAULT_HEADERS, create_fastapi
from hagworm.extend.base import Utils
from hagworm.extend.logging import LogFileRotator

from setting import Config
from service import DataSource
from controller import router


app = create_fastapi(
    log_level=Config.LogLevel,
    log_file_path=Config.LogFilePath,
    log_file_rotation=LogFileRotator.make(Config.LogFileSplitSize, Config.LogFileSplitTime),
    log_file_retention=Config.LogFileBackups,
    debug=Config.Debug,
    routes=router.routes,
    # on_startup=[DataSource.initialize],
    # on_shutdown=[DataSource.release],
)

app.add_middleware(
    CORSMiddleware,
    allow_origins=Config.AllowOrigins,
    allow_methods=Config.AllowMethods,
    allow_headers=Config.AllowHeaders,
    allow_credentials=Config.AllowCredentials,
)


if __name__ == r'__main__':

    Utils.log.warning(r'THE PRODUCTION ENVIRONMENT IS STARTED USING GUNICORN')

    uvicorn.run(app, port=Config.Port, log_config=None, headers=DEFAULT_HEADERS)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import typing
import socket
import asyncio
import contextlib
import contextvars

import grpc

from .server import GRPCServer, Router
from .server import UNARY_UNARY_MODE, UNARY_STREAM_MODE, STREAM_UNARY_MODE, STREAM_STREAM_MODE
from .client import GRPCClient

from ...extend.asyncio.base import Utils
from ...extend.asyncio.command import MainProcess, SubProcess


def get_metadata(remote_context: grpc.aio.ServicerContext) -> grpc.aio.Metadata:

    metadata = grpc.aio.Metadata((r'remote-peer', remote_context.peer()))

    for _key, _value in remote_context.invocation_metadata():
        metadata.add(_key, _value)

    return metadata


@contextlib.contextmanager
def grpc_context(
        remote_context: grpc.aio.ServicerContext,
        server_stream_stub: typing.Union[grpc.aio.UnaryStreamCall, grpc.aio.StreamStreamCall] = None
):

    remote_peer = remote_context.peer()

    Utils.log.info(f'grpc proxy open stream: {remote_peer}')

    try:
        yield
    except grpc.aio.UsageError:
        pass
    except Exception as err:
        Utils.log.exception(str(err))
    finally:
        if server_stream_stub is not None:
            server_stream_stub.cancel()

    Utils.log.info(f'grpc proxy close stream: {remote_peer}')


class ProxyUnaryStream:

    def __init__(
            self, server_proxy: GRPCClient, server_method: str,
            remote_request: bytes, remote_context: grpc.aio.ServicerContext
    ):

        self._server_proxy: GRPCClient = server_proxy

        self._server_stream_stub: grpc.aio.UnaryStreamCall = self._server_proxy.unary_stream(
            server_method, remote_request,
            metadata=get_metadata(remote_context)
        )

        self._remote_context: grpc.aio.ServicerContext = remote_context

        self._stream_task: asyncio.Task = asyncio.create_task(self._do_stream_task())

    async def _do_stream_task(self):

        with grpc_context(self._remote_context, self._server_stream_stub):
            async for _message in self._server_stream_stub:
                await self._remote_context.write(_message)

    async def join(self) -> typing.Any:
        return await self._stream_task


class ProxyStreamUnary:

    def __init__(
            self, server_proxy: GRPCClient, server_method: str,
            remote_request: typing.AsyncIterator, remote_context: grpc.aio.ServicerContext
    ):

        self._server_proxy: GRPCClient = server_proxy

        self._server_stream_stub: grpc.aio.StreamUnaryCall = self._server_proxy.stream_unary(
            server_method,
            metadata=get_metadata(remote_context)
        )

        self._remote_request: typing.AsyncIterator = remote_request
        self._remote_context: grpc.aio.ServicerContext = remote_context

        self._stream_task: asyncio.Task = asyncio.create_task(self._do_stream_task())

    async def _do_stream_task(self):

        with grpc_context(self._remote_context):
            async for _message in self._remote_request:
                await self._server_stream_stub.write(_message)

    async def join(self) -> typing.Any:
        return await self._server_stream_stub


class ProxyStreamStream:

    def __init__(
            self, server_proxy: GRPCClient, server_method: str,
            remote_request: typing.AsyncIterator, remote_context: grpc.aio.ServicerContext
    ):

        self._server_proxy: GRPCClient = server_proxy
        self._server_stream_stub: grpc.aio.StreamStreamCall = self._server_proxy.stream_stream(
            server_method,
            metadata=get_metadata(remote_context)
        )

        self._remote_request: typing.AsyncIterator = remote_request
        self._remote_context: grpc.aio.ServicerContext = remote_context

        self._stream_task: asyncio.Task = asyncio.create_task(self._do_stream_task())

    async def _read_upstream(self) -> bytes:

        try:
            return await anext(self._remote_request)
        except StopAsyncIteration:
            pass
        except Exception as err:
            Utils.log.error(f'read upstream error: {str(err)}')
            raise asyncio.CancelledError()

    async def _read_downstream(self) -> bytes:

        try:
            return await self._server_stream_stub.read()
        except Exception as err:
            Utils.log.error(f'read downstream error: {str(err)}')
            raise asyncio.CancelledError()

    async def _do_stream_task(self):

        with grpc_context(self._remote_context, self._server_stream_stub):

            upstream_task = asyncio.create_task(self._read_upstream())
            downstream_task = asyncio.create_task(self._read_downstream())

            while True:

                _tasks_done, _ = await asyncio.wait(
                    [upstream_task, downstream_task],
                    return_when=asyncio.FIRST_COMPLETED
                )

                for _task in _tasks_done:

                    _message = await _task

                    if _message == grpc.aio.EOF:
                        break

                    if _task is upstream_task:
                        await self._server_stream_stub.write(_message)
                        upstream_task = asyncio.create_task(self._read_upstream())
                    elif _task is downstream_task:
                        await self._remote_context.write(_message)
                        downstream_task = asyncio.create_task(self._read_downstream())

    async def join(self) -> typing.Any:
        return await self._stream_task


class ProxyInterceptor(grpc.aio.ServerInterceptor):

    def __init__(self, client: GRPCClient, routers: typing.List[Router]):

        self._method_context: contextvars.ContextVar = contextvars.ContextVar[str](r'grpc_method_context', default=None)

        self._client: GRPCClient = client

        self._handlers: typing.Dict[str, grpc.RpcMethodHandler] = {}

        self._init_handlers(routers)

    def _init_handlers(self, routers: typing.List[Router]):

        for router in routers:

            _service = router.service

            for _func, _mode in router.handlers:

                _method = f'/{_service}/{_func.__name__}'

                if _mode is UNARY_UNARY_MODE:
                    self._handlers[_method] = grpc.unary_unary_rpc_method_handler(self._unary_unary_handler)
                elif _mode is UNARY_STREAM_MODE:
                    self._handlers[_method] = grpc.unary_stream_rpc_method_handler(self._unary_stream_handler)
                elif _mode is STREAM_UNARY_MODE:
                    self._handlers[_method] = grpc.stream_unary_rpc_method_handler(self._stream_unary_handler)
                elif _mode is STREAM_STREAM_MODE:
                    self._handlers[_method] = grpc.stream_stream_rpc_method_handler(self._stream_stream_handler)

    async def _unary_unary_handler(self, request: bytes, context: grpc.aio.ServicerContext) -> typing.Any:

        method = self._method_context.get()

        return await self._client.unary_unary(
            method, request,
            metadata=get_metadata(context)
        )

    async def _unary_stream_handler(self, request: bytes, context: grpc.aio.ServicerContext) -> typing.Any:

        method = self._method_context.get()

        return await ProxyUnaryStream(self._client, method, request, context).join()

    async def _stream_unary_handler(self, request: typing.AsyncIterator, context: grpc.aio.ServicerContext) -> typing.Any:

        method = self._method_context.get()

        return await ProxyStreamUnary(self._client, method, request, context).join()

    async def _stream_stream_handler(self, request: typing.AsyncIterator, context: grpc.aio.ServicerContext) -> typing.Any:

        method = self._method_context.get()

        return await ProxyStreamStream(self._client, method, request, context).join()

    async def intercept_service(
            self,
            continuation: typing.Callable[[grpc.HandlerCallDetails], typing.Awaitable[grpc.RpcMethodHandler]],
            handler_call_details: grpc.HandlerCallDetails
    ) -> grpc.RpcMethodHandler:

        method = getattr(handler_call_details, r'method')

        self._method_context.set(method)

        if method in self._handlers:
            return self._handlers.get(method)
        else:
            return await continuation(handler_call_details)


class GRPCProxy:

    def __init__(
            self, address: typing.Union[str, typing.Tuple[str, int]], routers: typing.List[Router],
            grpc_family: int = socket.AF_INET, grpc_server_credentials: typing.Optional[grpc.ServerCredentials] = None
    ):

        self._routers: typing.List[Router] = routers

        self._grpc_client: typing.Optional[GRPCClient] = None
        self._grpc_server: typing.Optional[GRPCServer] = None

        self._grpc_address: typing.Union[str, typing.Tuple[str, int]] = address
        self._grpc_family: int = grpc_family
        self._grpc_server_credentials: typing.Optional[grpc.ServerCredentials] = grpc_server_credentials

    async def _register(self, request: typing.AsyncIterator, context: grpc.aio.ServicerContext):

        target = dict(context.invocation_metadata()).get(r'worker_entry_point')

        if target:
            context.add_done_callback(lambda _: Utils.create_task(self._grpc_client.clear_targets([target])))
        else:
            await context.abort(grpc.StatusCode.CANCELLED)

        await self._grpc_client.append_target(target)

        Utils.log.success(f'grpc proxy connected worker: {target}')

        with Utils.suppress(grpc.RpcError):
            async for _message in request:
                Utils.log.warning(f'grpc proxy received message: {_message}')

    async def open(self):

        self._grpc_client: GRPCClient = GRPCClient(
            request_serializer=None,
            response_deserializer=None
        )

        self._grpc_server: GRPCServer = GRPCServer(
            interceptors=[ProxyInterceptor(self._grpc_client, self._routers)],
            request_deserializer=None,
            response_serializer=None
        )

        self._grpc_server.register(r'', [self._register], mode=STREAM_STREAM_MODE)

        await self._grpc_server.start(self._grpc_address, self._grpc_family, self._grpc_server_credentials)

    async def close(self):

        await self._grpc_server.stop()
        await self._grpc_client.close()

    async def join(self):
        await self._grpc_server.wait()


class GRPCEntry:

    def __init__(self, proxy_address: typing.Union[str, typing.Tuple[str, int]], routers: typing.List[Router]):

        self._proxy_address: str = self._decode_address(proxy_address)
        self._server_address: str = f'/tmp/grpc_worker_{Utils.getpid()}.sock'

        self._grpc_server: typing.Optional[GRPCServer] = None
        self._grpc_routers: typing.List[Router] = routers

        self._proxy_client: typing.Optional[GRPCClient] = None
        self._proxy_client_stub: typing.Optional[grpc.aio.StreamStreamCall] = None

    @staticmethod
    def _decode_address(proxy_address: typing.Union[str, typing.Tuple[str, int]]) -> str:

        if isinstance(proxy_address, str):
            proxy_address = Utils.split_str(proxy_address, r':')

        if proxy_address[0] == r'0.0.0.0':
            proxy_address =(r'localhost', proxy_address[1])

        return Utils.join_str(proxy_address, r':')

    async def start(self):

        if os.path.exists(self._server_address):
            os.remove(self._server_address)

        self._grpc_server = GRPCServer()

        for _router in self._grpc_routers:
            self._grpc_server.bind_router(_router)

        await self._grpc_server.start(self._server_address, family=socket.AF_UNIX)

        self._proxy_client = GRPCClient()
        await self._proxy_client.open([self._proxy_address])

        self._proxy_client_stub = self._proxy_client.stream_stream(
            r'//_register',
            metadata=grpc.aio.Metadata((r'worker_entry_point', r'unix:' + self._server_address))
        )

    async def stop(self):
        await self._proxy_client.close()
        await self._grpc_server.stop()

    async def join(self):

        with Utils.suppress(grpc.RpcError):
            async for _message in self._proxy_client_stub:
                Utils.log.warning(f'grpc worker received message: {_message}')

        await self._grpc_server.wait()


class GRPCMainProcess(MainProcess):

    def __init__(
            self, address: typing.Union[str, typing.Tuple[str, int]], routers: typing.List[Router],
            worker: typing.Type[r'GRPCWorker'], sub_process_num: int, *,
            grpc_family: int = socket.AF_INET, grpc_server_credentials: typing.Optional[grpc.ServerCredentials] = None,
            cpu_affinity: bool = False, join_timeout: int = 10
    ):

        super().__init__(
            worker.create, sub_process_num,
            cpu_affinity=cpu_affinity, join_timeout=join_timeout,
            proxy_address=address, routers=routers
        )

        self._grpc_proxy: GRPCProxy = GRPCProxy(address, routers, grpc_family, grpc_server_credentials)

    async def initialize(self):
        await self._grpc_proxy.open()

    async def release(self):
        await self._grpc_proxy.close()


class GRPCWorker(SubProcess):

    def __init__(self, process_num: int, proxy_address: typing.Union[str, typing.Tuple[str, int]], routers: typing.List[Router]):

        super().__init__(process_num)

        self._grpc_entry: GRPCEntry = GRPCEntry(proxy_address, routers)

    async def initialize(self):
        await self._grpc_entry.start()

    async def release(self):
        await self._grpc_entry.stop()

    async def _execute(self):
        await self._grpc_entry.join()



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing

from pydantic import AfterValidator
from typing_extensions import Annotated

from ...extend import validator
from ...extend.base import Utils


def _IDCardType(val: str) -> str:

    if not Utils.identity_card(val):
        raise ValueError(r'value is not a valid identity card')

    return val

IDCardType = Annotated[str, AfterValidator(_IDCardType)]


def _BankCardType(val: str) -> str:

    if not Utils.luhn_valid(val):
        raise ValueError(r'value is not a valid bank card')

    return val

BankCardType = Annotated[str, AfterValidator(_BankCardType)]


def _UUIDType(val: str) -> str:

    if not validator.uuid(val):
        raise ValueError(r'value is not a valid uuid')

    return val

UUIDType = Annotated[str, AfterValidator(_UUIDType)]


def _DateType(val: str) -> typing.Any:

    try:
        val = Utils.date_parse(val)
    except Exception:
        raise ValueError(r'value is not a valid date')

    return val

DateType = Annotated[str, AfterValidator(_DateType)]


def _JsonType(val: str) -> typing.Any:

    try:
        val = Utils.json_decode(val)
    except Exception:
        raise ValueError(r'value is not a valid json')

    return val

JsonType = Annotated[str, AfterValidator(_JsonType)]


def _IntListType(val: str) -> typing.List[int]:

    try:
        val = Utils.split_int(val)
    except Exception:
        raise ValueError(r'value is not a valid int list')

    return val

IntListType = Annotated[str, AfterValidator(_IntListType)]


def _FloatListType(val: str) -> typing.List[float]:

    try:
        val = Utils.split_float(val)
    except Exception:
        raise ValueError(r'value is not a valid float list')

    return val

FloatListType = Annotated[str, AfterValidator(_FloatListType)]


def _ASCVisibleType(val: str) -> str:

    if not validator.asc_visible(val):
        raise ValueError(r'value is not a valid asc visible')

    return val

ASCVisibleType = Annotated[str, AfterValidator(_ASCVisibleType)]


def _EmailType(val: str) -> str:

    if not validator.email(val):
        raise ValueError(r'value is not a valid email')

    return val

EmailType = Annotated[str, AfterValidator(_EmailType)]


def _DomainType(val: str) -> str:

    if not validator.domain(val):
        raise ValueError(r'value is not a valid domain')

    return val

DomainType = Annotated[str, AfterValidator(_DomainType)]


def _URLType(val: str) -> str:

    if not validator.url(val):
        raise ValueError(r'value is not a valid url')

    return val

URLType = Annotated[str, AfterValidator(_URLType)]


def _MacAddrType(val: str) -> str:

    if not validator.mac_addr(val):
        raise ValueError(r'value is not a valid mac address')

    return val

MacAddrType = Annotated[str, AfterValidator(_MacAddrType)]


def _IPvAnyType(val: str) -> str:

    if not validator.ipv4(val) and not validator.ipv6(val):
        raise ValueError(r'value is not a valid ip address')

    return val

IPvAnyType = Annotated[str, AfterValidator(_IPvAnyType)]


def _IPv4Type(val: str) -> str:

    if not validator.ipv4(val):
        raise ValueError(r'value is not a valid ipv4 address')

    return val

IPv4Type = Annotated[str, AfterValidator(_IPv4Type)]


def _IPv4CidrType(val: str) -> str:

    if not validator.ipv4_cidr(val):
        raise ValueError(r'value is not a valid ipv4 address')

    return val

IPv4CidrType = Annotated[str, AfterValidator(_IPv4CidrType)]


def _IPv6Type(val: str) -> str:

    if not validator.ipv6(val):
        raise ValueError(r'value is not a valid ipv6 address')

    return val

IPv6Type = Annotated[str, AfterValidator(_IPv6Type)]


def _IPv6CidrType(val: str) -> str:

    if not validator.ipv6_cidr(val):
        raise ValueError(r'value is not a valid ipv6 address')

    return val

IPv6CidrType = Annotated[str, AfterValidator(_IPv6CidrType)]



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import signal
import typing
import asyncio
import copy
import functools
import logging

from threading import Event, Thread as _Thread
from concurrent.futures import ThreadPoolExecutor

from ..interface import RunnableInterface, TaskInterface


class WaitForever(asyncio.Future):

    def __init__(self):

        super().__init__()

        signal.signal(signal.SIGINT, self._exit)
        signal.signal(signal.SIGTERM, self._exit)

    def _exit(self, *_):

        if not self.cancelled():
            self.cancel()


class FutureWithTimeout(asyncio.Future):
    """带超时功能的Future
    """

    def __init__(self, delay: float, default: typing.Any = None):

        super().__init__()

        self._timeout_handle = asyncio.get_event_loop().call_later(
            delay,
            self.set_result,
            default
        )

        self.add_done_callback(self._clear_timeout)

    def _clear_timeout(self, *_):

        if self._timeout_handle is not None:
            self._timeout_handle.cancel()
            self._timeout_handle = None


class FutureWithCoroutine(asyncio.Future):
    """Future实例可以被多个协程await，本类实现Future和Coroutine的桥接
    """

    def __init__(self, coroutine: typing.Coroutine):

        super().__init__()

        self._lock = asyncio.Lock()
        self._coroutine: typing.Coroutine = coroutine

    def __await__(self) -> typing.Any:
        return (yield from self._run_coroutine().__await__())

    async def _run_coroutine(self):

        async with self._lock:

            if not self.done():

                try:
                    self.set_result(
                        await self._coroutine
                    )
                except Exception as err:
                    super().cancel()
                    raise err

        return copy.deepcopy(self.result())

    def cancel(self, *args, **kwargs):

        super().cancel()

        self._coroutine.close()


class Thread(_Thread):
    """带强制停止功能的线程，如非必要请勿使用
    """
    
    def __init__(self, group=None, target=None, name=None, args=(), kwargs=None):

        super().__init__(group, target, name, args, kwargs, daemon=True)

        self._exit_event: Event = Event()

    def stop(self, timeout: int = None):

        self._exit_event.set()

        if timeout is not None:
            self.join(timeout)

    def is_stopped(self) -> bool:

        return self._exit_event.is_set()


class ThreadPool(RunnableInterface):
    """线程池，桥接线程与协程
    """

    def __init__(self, max_workers: typing.Optional[int] = None):

        self._executor = ThreadPoolExecutor(max_workers)

    async def run(self, _callable: typing.Callable, *args, **kwargs):
        """线程转协程，不支持协程函数
        """

        loop = asyncio.events.get_event_loop()

        if kwargs:

            return await loop.run_in_executor(
                self._executor,
                functools.partial(
                    _callable,
                    *args,
                    **kwargs
                )
            )

        else:

            return await loop.run_in_executor(
                self._executor,
                _callable,
                *args,
            )


class ThreadWorker:
    """通过线程转协程实现普通函数非阻塞的装饰器
    """

    def __init__(self, max_workers: typing.Optional[int] = None):

        self._thread_pool = ThreadPool(max_workers)

    def __call__(self, func: typing.Callable):

        @functools.wraps(func)
        def _wrapper(*args, **kwargs):
            return self._thread_pool.run(func, *args, **kwargs)

        return _wrapper


class SubProcess(TaskInterface):
    """子进程管理，通过command方式启动子进程
    """

    @classmethod
    async def create(cls, program, *args, stdin=None, stdout=None, stderr=None, **kwargs):

        inst = cls(program, *args, stdin=stdin, stdout=stdout, stderr=stderr, **kwargs)
        await inst.start()

        return inst

    def __init__(
            self, program: str, *args,
            stdin: int = asyncio.subprocess.DEVNULL,
            stdout: int = asyncio.subprocess.DEVNULL,
            stderr: int= asyncio.subprocess.DEVNULL,
            **kwargs
    ):

        self._program: str = program
        self._args: typing.Tuple = args
        self._kwargs: typing.Dict = kwargs

        self._stdin: int = stdin
        self._stdout: int = stdout
        self._stderr: int = stderr

        self._process: typing.Optional['asyncio.subprocess.Process'] = None
        self._process_id: typing.Optional[int] = None

    @property
    def pid(self) -> int:
        return self._process_id

    @property
    def process(self) -> 'asyncio.subprocess.Process':
        return self._process

    @property
    def stdin(self) -> 'asyncio.streams.StreamWriter':
        return self._process.stdin

    @property
    def stdout(self) -> 'asyncio.streams.StreamReader':
        return self._process.stdout

    @property
    def stderr(self) -> 'asyncio.streams.StreamReader':
        return self._process.stderr

    def create_stdout_task(self) -> asyncio.Task:
        return asyncio.create_task(self._log_stdout())

    async def _log_stdout(self):

        stream = self._process.stdout

        while self._process.returncode is None:

            content = await stream.readline()

            if content:
                content = content.decode().strip()
                if content:
                    logging.info(content)
            else:
                break

    def create_stderr_task(self) -> asyncio.Task:

        return asyncio.create_task(self._log_stderr())

    async def _log_stderr(self):

        stream = self._process.stderr

        while self._process.returncode is None:

            content = await stream.readline()

            if content:
                content = content.decode().strip()
                if content:
                    logging.error(content)
            else:
                break

    def is_running(self) -> bool:
        return self._process is not None and self._process.returncode is None

    async def start(self) -> bool:

        if self.is_running():
            return False

        self._process = await asyncio.create_subprocess_exec(
            self._program, *self._args,
            stdin=self._stdin,
            stdout=self._stdout,
            stderr=self._stderr,
            **self._kwargs
        )

        self._process_id = self._process.pid

        return True

    async def stop(self):

        if self.is_running():
            self._process.kill()
            await self._process.wait()

    def kill(self):

        if self.is_running():
            self._process.kill()

    async def wait(self, timeout: typing.Optional[float] = None):

        if not self.is_running():
            return

        try:
            await asyncio.wait_for(self._process.wait(), timeout=timeout)
        except Exception as err:
            logging.error(err)
        finally:
            await self.stop()



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import sys
import asyncio

os.chdir(os.path.dirname(__file__))
sys.path.insert(0, os.path.abspath(r'../'))

from hagworm.extend.asyncio.base import Utils
from hagworm.extend.asyncio.socket import AsyncTcpServer


async def connection_handle(reader: asyncio.StreamReader, writer: asyncio.StreamWriter):

    client_info = writer.transport.get_extra_info(r'socket').getpeername()

    Utils.log.info(f'create connection: {client_info}')

    try:

        while True:

                request = await reader.readline()

                if not request:
                    break

                writer.write(request)

                await writer.drain()

    except Exception as err:

        Utils.log.error(err)

    finally:

        if not writer.is_closing():
            writer.close()
            await writer.wait_closed()

    Utils.log.info(f'lost connection: {client_info}')


def main():

    server = AsyncTcpServer(connection_handle, (r'0.0.0.0', 8080))
    server.run()


if __name__ == r'__main__':

    main()



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing
import asyncio
import grpc
import msgpack

from abc import abstractmethod

from ...extend.struct import RoundRobin
from ...extend.asyncio.base import Utils


CHANNEL_USABLE_STATE = (grpc.ChannelConnectivity.READY, grpc.ChannelConnectivity.IDLE)


class GRPCClient:

    def __init__(
            self, *,
            credentials: typing.Optional[grpc.ChannelCredentials] = None,
            options: typing.Optional[grpc.aio.ChannelArgumentType] = None,
            compression: typing.Optional[grpc.Compression] = None,
            interceptors: typing.Optional[typing.Sequence[grpc.aio.ClientInterceptor]] = None,
            request_serializer: typing.Optional[typing.Callable] = msgpack.dumps,
            response_deserializer: typing.Optional[typing.Callable] = msgpack.loads
    ):

        self._credentials: typing.Optional[grpc.ChannelCredentials] = credentials
        self._options: typing.Optional[grpc.aio.ChannelArgumentType] = options
        self._compression: typing.Optional[grpc.Compression] = compression
        self._interceptors: typing.Optional[typing.Sequence[grpc.aio.ClientInterceptor]] = interceptors

        self._request_serializer: typing.Optional[typing.Callable] = request_serializer
        self._response_deserializer: typing.Optional[typing.Callable] = response_deserializer

        self._channels: RoundRobin[grpc.aio.Channel] = RoundRobin()

    def _make_channel(self, target: str) -> grpc.aio.Channel:

        if self._credentials is None:
            return grpc.aio.insecure_channel(
                target, self._options, self._compression, self._interceptors
            )
        else:
            return grpc.aio.secure_channel(
                target, self._credentials, self._options, self._compression, self._interceptors
            )

    def _make_channels(self, targets: typing.List[str]) -> typing.Dict[str, grpc.aio.Channel]:
        return {_target: self._make_channel(_target) for _target in set(targets)}

    async def append_target(self, target: str):

        _channel = self._channels.append(target, self._make_channel(target))

        if _channel is not None:
            await _channel.close()

    async def reset_targets(self, targets: typing.List[str]):

        _channels = self._make_channels(targets)

        for _, _channel in self._channels.reset(_channels).items():
            await _channel.close()

    async def clear_targets(self, keys: typing.Optional[typing.List[str]] = None):

        for _, _channel in self._channels.clear(keys).items():
            await _channel.close()

    async def open(self, targets: typing.List[str]):
        await self.reset_targets(targets)

    async def close(self):
        await self.clear_targets()

    def get_channel(self) -> grpc.aio.Channel:

        for _ in range(len(self._channels)):

            _, channel = self._channels.get()

            if channel.get_state() in CHANNEL_USABLE_STATE:
                break

        else:

            raise ConnectionAbortedError()

        return channel

    def unary_unary(
            self, method: str, call_params: typing.Union[bytes, typing.Dict],
            *, timout: typing.Optional[float] = None, metadata: typing.Optional[grpc.aio.Metadata] = None
    ) -> grpc.aio.UnaryUnaryCall:

        channel = self.get_channel()

        return channel.unary_unary(
            method,
            request_serializer=self._request_serializer,
            response_deserializer=self._response_deserializer,
        )(call_params, timeout=timout, wait_for_ready=True, metadata=metadata)

    def unary_stream(
            self, method: str, call_params: typing.Union[bytes, typing.Dict],
            *, timout: typing.Optional[float] = None, metadata: typing.Optional[grpc.aio.Metadata] = None
    ) -> grpc.aio.UnaryStreamCall:

        channel = self.get_channel()

        return channel.unary_stream(
            method,
            request_serializer=self._request_serializer,
            response_deserializer=self._response_deserializer,
        )(call_params, timeout=timout, wait_for_ready=True, metadata=metadata)

    def stream_unary(
            self, method: str,
            *, timout: typing.Optional[float] = None, metadata: typing.Optional[grpc.aio.Metadata] = None
    ) -> grpc.aio.StreamUnaryCall:

        channel = self.get_channel()

        return channel.stream_unary(
            method,
            request_serializer=self._request_serializer,
            response_deserializer=self._response_deserializer,
        )(timeout=timout, wait_for_ready=True, metadata=metadata)

    def stream_stream(
            self, method: str,
            *, timout: typing.Optional[float] = None, metadata: typing.Optional[grpc.aio.Metadata] = None
    ) -> grpc.aio.StreamStreamCall:

        channel = self.get_channel()

        return channel.stream_stream(
            method,
            request_serializer=self._request_serializer,
            response_deserializer=self._response_deserializer,
        )(timeout=timout, wait_for_ready=True, metadata=metadata)


class RobustStreamClient:

    def __init__(self, client: GRPCClient, method: str):

        self._grpc_client: GRPCClient = client

        self._stream_method: str = method

        self._stream_stub: typing.Optional[grpc.aio.StreamStreamCall] = None
        self._stream_task: typing.Optional[asyncio.Task] = None

    async def _do_stream_task(self):

        try:
            async for _message in self._stream_stub:
                await self.on_message(_message)
        except Exception as err:
            Utils.log.error(str(err))

    async def join(self) -> typing.Any:
        return await self._stream_task

    async def connect(self, *, timout: typing.Optional[float] = None, metadata: typing.Optional[grpc.aio.Metadata] = None) -> bool:

        try:

            self.close()

            self._stream_stub = self._grpc_client.stream_stream(self._stream_method, timout=timout, metadata=metadata)
            self._stream_stub.add_done_callback(self.on_close)

            await self.on_connect()

            self._stream_task = asyncio.create_task(self._do_stream_task())

            return True

        except Exception as err:

            Utils.log.error(str(err))

            return False

    def close(self):

        if self._stream_task is not None and not self._stream_task.done() and not self._stream_task.cancelling():
            self._stream_task.cancel()

        if self._stream_stub is not None and not self._stream_stub.done():
            self._stream_stub.cancel()

        self._stream_task = None
        self._stream_stub = None

    async def read(self) -> typing.Any:

        data = await self._stream_stub.read()

        if data != grpc.aio.EOF:
            return data
        else:
            raise ConnectionAbortedError()

    async def send(self, message: typing.Any):
        await self._stream_stub.write(message)

    @abstractmethod
    async def on_message(self, message: typing.Any):
        """
        接收消息回调
        """

    @abstractmethod
    async def on_connect(self):
        """
        与服务端通信连接成功后回调
        """

    @abstractmethod
    def on_close(self, stub: grpc.aio.StreamStreamCall):
        """
        会话关闭回调
        """



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import typing
import dataclasses
import json
import yaml

from configparser import RawConfigParser

from .base import Utils


class HostType(str):

    @classmethod
    def decode(cls, val: str) -> typing.Tuple[str, int]:

        if val:
            host, port = val.split(r':', 2)
            return host.strip(), int(port.strip())


class JsonType(str):

    @classmethod
    def decode(cls, val: str) -> typing.Any:

        if val:
            return Utils.json_decode(val)


class StrListType(str):

    @classmethod
    def decode(cls, val: str) -> typing.List[str]:
        return Utils.split_str(val)


class IntListType(str):

    @classmethod
    def decode(cls, val: str) -> typing.List[int]:
        return Utils.split_int(val)


class FloatListType(str):

    @classmethod
    def decode(cls, val: str) -> typing.List[float]:
        return Utils.split_float(val)


class Field:

    __slots__ = [r'section', r'default']

    def __init__(self, section: str, default: typing.Any = None):

        self.section: str = section
        self.default: typing.Any = default


class ConfigureMetaclass(type):
    """配置类元类，增加dataclass修饰
    """

    def __new__(mcs, name, bases, attrs):
        # noinspection PyTypeChecker
        return dataclasses.dataclass(init=False)(
            type.__new__(mcs, name, bases, attrs)
        )


class KeySection(typing.NamedTuple):

    key: str
    type: typing.Any
    section: str
    default: typing.Any


class ConfigureBase(metaclass=ConfigureMetaclass):
    """配置类
    """

    __slots__ = [r'__dataclass_fields__', r'_parser', r'_key_section']

    def __init__(self):

        super().__init__()

        self._parser = RawConfigParser()

        self._key_section: typing.Dict[str, KeySection] = {
            f'{_field.default.section}_{_key}': KeySection(
                _key,
                _field.type,
                _field.default.section,
                _field.default.default,
            )
            for _key, _field in self.__dataclass_fields__.items()
        }

    def _load_options(self):

        for _key_section in self._key_section.values():

            # 优先处理环境变量
            _env_key = f'{_key_section.section}_{_key_section.key}'.upper()
            _env_val = os.getenv(_env_key, None)

            if _env_val is not None:
                self._parser.set(_key_section.section, _key_section.key, _env_val)
                Utils.log.info(f'load environment variable {_env_key}: {_env_val}')

            if _key_section.type is str:
                _val = self._parser.get(_key_section.section, _key_section.key)
            elif _key_section.type is int:
                _val = self._parser.getint(_key_section.section, _key_section.key)
            elif _key_section.type is float:
                _val = self._parser.getfloat(_key_section.section, _key_section.key)
            elif _key_section.type is bool:
                _val = self._parser.getboolean(_key_section.section, _key_section.key)
            else:
                _val = _key_section.type.decode(self._parser.get(_key_section.section, _key_section.key))

            self.__setattr__(_key_section.key, _val)

    def _clear_options(self):

        self._parser.clear()

        _config = {}

        for _key_section in self._key_section.values():
            if _key_section.default is not None:
                _config.setdefault(_key_section.section, {})[_key_section.key] = _key_section.default

        if _config:
            self._parser.read_dict(_config)

    def to_dict(self, section=None) -> typing.Dict:

        if section is None:
            return {key: dict(val) for key, val in self._parser.items()}
        else:
            return {key: val for key, val in self._parser.items(section)}

    def to_yaml(self, section=None) -> str:
        return yaml.dump(self.to_dict(section))


class Configure(ConfigureBase):
    """配置类
    """

    def get_option(self, section: str, option: str) -> typing.Any:

        return self._parser.get(section, option)

    def get_options(self, section: str) -> typing.Dict:

        parser = self._parser

        options = {}

        for option in parser.options(section):
            options[option] = parser.get(section, option)

        return options

    def set_options(self, section: str, **options: str):

        if not self._parser.has_section(section):
            self._parser.add_section(section)

        for option, value in options.items():
            self._parser.set(section, option, value)

        self._load_options()

    def read(self, path: str, encoding: str = r'utf-8'):

        self._clear_options()

        self._parser.read(path, encoding=encoding)

        self._load_options()

    def read_env(self):

        self._clear_options()

        self._load_options()

    def read_str(self, val: str):

        self._clear_options()

        self._parser.read_string(val)

        self._load_options()

    def read_dict(self, val: typing.Dict):

        self._clear_options()

        self._parser.read_dict(val)

        self._load_options()

    def read_json(self, path: str, encoding: str = r'utf-8'):

        self._clear_options()

        with open(path, encoding=encoding) as fp:
            self._parser.read_dict(json.loads(fp.read()))

        self._load_options()

    def read_yaml(self, path: str, encoding: str = r'utf-8'):

        self._clear_options()

        with open(path, encoding=encoding) as fp:
            self._parser.read_dict(yaml.load(fp.read(), yaml.Loader))

        self._load_options()



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing

from fastapi.responses import UJSONResponse

from ...extend.trace import get_trace_id
from ...extend.error import Ignore


class Response(UJSONResponse):

    def __init__(self, content: typing.Any = None, status_code: int = 200, *args, **kwargs):

        self._trace_id: str = get_trace_id()

        super().__init__(content, status_code, *args, **kwargs)

    def render(self, content: typing.Any):

        return UJSONResponse.render(
            self,
            dict(code=0, data=content, trace_id=self._trace_id)
        )


class ErrorResponse(Response, Ignore):

    def __init__(self, code: int, content: typing.Any = None, status_code: int = 200, **kwargs):

        self._code: int = code

        Response.__init__(self, content, status_code, **kwargs)
        Ignore.__init__(self, self.body.decode(), layers=-1)

    def render(self, content: typing.Any):

        return UJSONResponse.render(
            self,
            dict(code=self._code, error=content, trace_id=self._trace_id)
        )
